GUI_WIDTH = 1000
GUI_HEIGHT = 400
WIDGET_VIEW_ITEMS = 8
DEFAULT_VIEW_FORMAT = "Table"
MAX_LOG_SIZE = 20
MAX_MESSAGES = 3
META_FIELDS = ["_version", "_latest_version", "_etag", "_created"]
SHOW_INDICATOR = True
SIZING_MODE = "stretch_width"
DEBUG = False

import param
import json
import os


class ConfigParameter(param.Parameter):

    __slots__ = ["env_prefix", "klass"]

    def __init__(self, klass, env_prefix="", **kwargs):
        super().__init__(**kwargs)
        self.env_prefix = env_prefix
        self.klass = klass
       
    def _set_names(self, attrib_name):
        env_name = attrib_name.upper()
        env_name = self.env_prefix.upper() + "_" + env_name
        if os.getenv(env_name, ""):
            env = os.getenv(env_name, "")
            try:
                env = json.loads(env)
            except Exception as e:
                pass
            self.default = self.klass(env)
        super()._set_names(attrib_name)
        

class Config(param.Parameterized):
    GUI_MAX_WIDTH = ConfigParameter(int, env_prefix="eve_panel")
    GUI_MAX_HEIGHT = ConfigParameter(int, env_prefix="eve_panel")
    SHOW_INDICATOR = ConfigParameter(bool, env_prefix="eve_panel", default=True)
    SIZING_MODE = ConfigParameter(str, env_prefix="eve_panel", default="stretch_width")
    DEBUG = ConfigParameter(bool, env_prefix="eve_panel", default=False)
    GUI_WIDTH = 1000
    GUI_HEIGHT = 400
    WIDGET_VIEW_ITEMS = 8
    DEFAULT_VIEW_FORMAT = "Table"
    MAX_LOG_SIZE = 20
    MAX_MESSAGES = 3
    META_FIELDS = ["_version", "_latest_version", "_etag", "_created"]
    SHOW_INDICATOR = True
    DEFAULT_TIMEOUT = 20
    IGNORE_ERRORS = False
    
    OAUTH_DOMAIN = ConfigParameter(str, env_prefix="eve_panel", default="http://localhost/oauth")
    OAUTH_CERT_PATH = ConfigParameter(str, env_prefix="eve_panel", default="/.well-know/certs")
    OAUTH_CODE_PATH = ConfigParameter(str, env_prefix="eve_panel", default="/device/code")
    OAUTH_TOKEN_PATH = ConfigParameter(str, env_prefix="eve_panel", default="/token")
    DEFAULT_CLIENT_ID = ConfigParameter(str, env_prefix="eve_panel", default="eve-panel")
    DEFAULT_AUDIENCE = ConfigParameter(str, env_prefix="eve_panel", default="")
    

config = Config()



import base64
import panel as pn
import param
import getpass
import logging
import httpx
import webbrowser

from .base import EveAuthBase

logger = logging.getLogger(__name__)


class Oauth2DeviceFlow(EveAuthBase):
    auth_server_uri = param.String(
        label="Authentication server",
        default=None,
        regex=r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))")
    code_request_path = param.String(default="/device/code", label="Code request path")
    token_request_path = param.String(default="token",label="Token request path")
    verification_path = param.String(default="authorize",label="Verification path")
    extra_headers = param.Dict(label="Extra headers")
    client_id = param.String(default="eve_client", label="Client ID")
    notify_email = param.String(default="", label="Notify Email")
    device_code = param.String()
    user_code = param.String()
    token = param.String()
    
    _cb = param.Parameter(default=None)
    
    def get_client(self):
        return httpx.Client(base_url=self.auth_server_uri, headers=self.extra_headers)
    
    @property
    def authorize_url(self):
        return f"{self.auth_server_uri.strip('/')}/{self.verification_path.strip('/')}?user_code={self.user_code}"
    
    def initiate_flow(self):
        data = {}
        with self.get_client() as client:
            try:
                resp = client.post(self.code_request_path,
                                    data={"client_id": self.client_id,
                                    "notify_email": self.notify_email})
                data = resp.json()
            except:
                pass
        self.device_code = data.get("device_code", "")
        self.user_code = data.get("user_code", "")
        interval = data.get("interval", 3)
        timeout = data.get("expires_in", 300)
        if not self.user_code:
            return
        if not self.device_code:
            return
        if self._cb is not None:
            self._cb.stop()
        self._cb = pn.state.add_periodic_callback(self.callback,
                                                period=interval*1000,
                                                count=int(timeout/interval)+1,
                                                )
        
    def authorize(self):
        return webbrowser.open(self.authorize_url)
    
    def authorize_link(self):
        html_pane = pn.pane.HTML(f"""
        <a id="log-in-link" class="nav-link" href="{self.authorize_url}" target="_blank">
         Authorize 
        </a>""",
        style={"cursor": "pointer",
                "border": "1px solid #ddd",
                "border-radius": "4px",
                "padding": "5px",})
        return html_pane
    
    def await_token(self):
        with self.get_client() as client:
            for _ in range(int(self.timeout/self.interval)+1):
                data = {}
                try:
                    resp = client.post(self.token_request_path, 
                                      data={"client_id": self.client_id,
                                           "device_code": self.device_code,})
                    data = resp.json()
                except:
                    pass
                token = data.get("access_token", "")
                if token:
                    self.token = token
                    break
                time.sleep(self.interval)
        return token

    def check_token(self):
        data = {}
        with self.get_client() as client:
            try:
                resp = client.post(self.token_request_path, 
                                  data={"client_id": self.client_id,
                                       "device_code": self.device_code,})
                data = resp.json()
            except:
                pass
        return data.get("access_token", "")
    
    def callback(self):
        token = self.check_token()
        if token and self._cb is not None:
            self.token = token
            self._cb.stop() 
            self._cb = None
            
    @param.depends("_cb", "token")            
    def credentials_view(self):
        init_flow_button = pn.widgets.Button(name="Generate",
                                             button_type="primary",
                                            width=70)
        init_flow_button.on_click(lambda event: self.initiate_flow())
        params = pn.Param(self.param, parameters=["token", "auth_server_uri",
                                                 "client_id","notify_email"],
                            widgets={"token": {"type":pn.widgets.TextAreaInput, 
                                               "width":300}},
                            max_width=300,
                            sizing_mode="stretch_width")
        buttons = pn.Row(init_flow_button)
        if self._cb is not None:
            buttons.append(self.authorize_link())
            buttons.append(pn.indicators.LoadingSpinner(value=True, width=20, height=20))
        return pn.Column(params, buttons, sizing_mode="stretch_width", width=300)
    
    def perform_flow(self):
        self.initiate_flow()
        return pn.Column(self.view)
        
    def get_headers(self):
        if self.token:
            return {"Authorization": f"Bearer {self.token}"}
        else:
            return {}

    def login(self):
        self.initiate_flow()
        self.authorize()
        token = self.await_token()
        return bool(token)

    @property    
    def logged_in(self):
        return bool(token)

    def logout(self):
        self.token = ""
        
    def make_panel(self):
        return pn.panel(self.credentials_view)
    
    def __getstate__(self):
        state = super().__getstate__()
        state.pop("_cb", None)
        return state













import itertools
import json
from io import BytesIO, StringIO
import time
import pandas as pd
import panel as pn
import param
import yaml
import typing
import math
import asyncio
import base64
from typing import Union, List, Dict, Tuple
import multiprocessing as mp
from tqdm.autonotebook import tqdm

from .settings import config as settings
from .eve_model import EveModelBase
from .field import SUPPORTED_SCHEMA_FIELDS, TYPE_MAPPING, Validator
from .session import DEFAULT_SESSION_CLASS, EveSessionBase
from .item import EveItem
from .page import EvePage, EvePageCache, PageZero
from .io import FILE_READERS, read_data_file
from .types import DASK_TYPE_MAPPING, COERCERS
from .utils import NumpyJSONENncoder, to_data_dict

try:
    from ruamel.yaml import YAML
    yaml = YAML()
    yaml.indent(mapping=4, sequence=4, offset=2)
except ImportError:
    import yaml

class EveResource(EveModelBase):
    """
    Interface for an Eve resource.
    Should be instantiated using an Eve resource definition:
        EveResource.from_resource_def(definition, name)

    Inheritance:
        EveModelBase:

    """
    session = param.ClassSelector(EveSessionBase, constant=True, precedence=-1)
    _paste_bin = param.ClassSelector(class_=pn.widgets.Ace, default=None, )
    _url = param.String(precedence=-1)
    _page_view_format = param.Selector(objects=["Table", "Widgets", "JSON"],
                                       default=settings.DEFAULT_VIEW_FORMAT,
                                       label="Display Format",
                                       precedence=1)
    upload_errors = param.List(default=[])
    _resource_def = param.Dict(default={}, constant=True, precedence=-1)
    schema = param.Dict(default={}, constant=True, precedence=-1)

    _cache = param.ClassSelector(class_=EvePageCache, default=EvePageCache())
    _item_class = param.ClassSelector(EveItem,
                                      is_instance=False,
                                      precedence=-1)
    # _upload_buffer = param.List(default=[], precedence=-1)
    _upload_buffer = param.String(default="", precedence=1)
    _file_buffer = param.ClassSelector(bytes)
    _filename = param.String()

    filters = param.Dict(default={}, doc="Filters")
    fields = param.List(default=[], precedence=1)
    sorting = param.List(default=[], precedence=1)

    items_per_page = param.Integer(default=100,
                                   label="Items per page",
                                   precedence=1)
    _prev_page_button = param.Action(lambda self: self.decrement_page(),
                                     label="\u23EA",
                                     precedence=1)
    page_number = param.Integer(default=0,
                                bounds=(0, None),
                                label="",
                                doc="Page number",
                                precedence=2)
    _reload_page_button = param.Action(lambda self: self.reload_page(),
                                     label="\u21BB",
                                     precedence=4)
    _next_page_button = param.Action(lambda self: self.increment_page(),
                                     label="\u23E9",
                                     precedence=3)
    _clear_cache_button = param.Action(lambda self: self.clear_cache(),
                                     label="\u2672",
                                     precedence=5)
    _plot_selection = param.String("None")
    _plot = param.Parameter(default=None)
    _progress = param.Parameter(default=None)
    _active = param.Boolean(default=False)

    @classmethod
    def from_resource_def(cls,
                          resource_def: dict,
                          resource_name: str,
                          session=None):
        """Generate a resource interface from a Eve resource definition. 

        Args:
            resource_def (dict): Eve resource definition
            resource_name (str): Name to use for this resource
            session (EveSession, optional): session to use. Defaults to None.
            

        Returns:
            EveResource: Interface to the remote resource.
        """
        resource = dict(resource_def)
        schema = resource.pop("schema")
        if session is None:
            session = DEFAULT_SESSION_CLASS()
        
        item = EveItem.from_schema(resource["item_title"].replace(" ", "_"),
                                   schema,
                                   resource["url"],
                                   session=session)
        item_class = item.__class__
        plots = list(resource.get("metadata", {}).get("plots", {}))
        params = dict(name=resource["resource_title"].replace(" ", "_"),
                      _url=resource["url"],
                      session=session,
                      _item_class=item_class,
                      _resource_def=resource,
                      schema=schema,
                      fields=list(schema))
        instance = cls(**params)
        if not instance.is_tabular:
            instance._page_view_format = "JSON"
        return instance

    def __getitem__(self, key):
        if key in self._cache:
            return self._cache[key]
        data = self.session.get("/".join([self._url, key]))
        if data:
            if self._file_fields:
                for k in self._file_fields:
                    try:
                        data[k] = base64.b64decode(data[k])
                    except:
                        pass
            item = self.make_item(**data)
            return item
        raise KeyError

    def __setitem__(self, key, value):
        self._cache[key] = value

    def __iter__(self):
        yield from self.values()

    def make_item(self, **kwargs):
        """Generate EveItem from key value pairs

        Returns:
            EveItem: EveItem instance that enforces schema of current resource.
        """
        return self._item_class(**kwargs, session=self.session)

    @property
    def projection(self):
        return {k: 1 for k in self.fields if k not in settings.META_FIELDS}
    
    @property
    def paste_bin(self):
        if self._paste_bin is None:
            self._paste_bin = pn.widgets.Ace(name="Paste Bin", value=json.dumps(self.schema, indent=4),
                                             language="json", width=int(self.max_width-50))
        return self._paste_bin
    
    @property
    def field_options(self):
        return list(self.schema) + settings.META_FIELDS

    @property
    def nitems(self):
        resp = self.get(where=self.filters,
                        projection={"_id": 1},
                        max_results=1,
                        page=1,
                        timeout=15,
                        )
        if "_meta" in resp:
            return int(resp["_meta"].get("total", 0))
        else:
            raise ConnectionError("Unable to connect to server.")
    
    @property
    def is_tabular(self):
        return not any([v["type"] in ['list', 'dict'] for v in self.schema.values()])

    @property
    def page_numbers(self):
        return list(range(1, math.ceil(self.nitems/self.items_per_page)+1))

    @property
    def sorting_options(self)->list:
        """Returns valid options for the sorting attribute.

        Returns:
            list: list of fields
        """
        return self.field_options + [f"-{col}" for col in self.field_options]

    @property
    def metadata(self) -> dict:
        """Resource metadata

        Returns:
            dict: Resource metadata
        """
        return self._resource_def.get("metadata", {})

    @property
    def _file_fields(self):
        return [k for k,v in self.schema.items() if v["type"]=="media"]
    
    @property    
    def _files(self):
        return {k: getattr(self, k) for k in self._file_fields}

    @param.depends("page_number")
    def gui_progress(self):
        if self._progress is None:
            self._progress = pn.indicators.Progress(max=1, active=self._active, align="end")
        if self.page_number > self._progress.max:
            self._progress.max = len(self.page_numbers)
        self._progress.value = self.page_number
        return self._progress

    def read_clipboard(self):
        """Read clipboard into uplaod buffer.
        """
        from pandas.io.clipboard import clipboard_get
        try:
            self.paste_bin.value = clipboard_get()
        except Exception as e:
            print(e)

    def read_file(self, f: typing.Union[typing.BinaryIO, str]=None, ext: str="csv"):
        """Read file into the upload buffer.

        Args:
            f (File, optional): file like object. Defaults to None.
            ext (str, optional): file extension. Defaults to "csv".

        Returns:
            list: documents read
        """
        if isinstance(f, str):
            f = open(f, "rb")
        data = read_data_file(f, ext)
        self.paste_bin.value = json.dumps(data, indent=4)
        return data

    @param.depends("_file_buffer", watch=True)
    def _read_file_buffer(self):
        sio = BytesIO(self._file_buffer)
        _,_, ext = self._filename.rpartition(".")
        self.read_file(sio, ext)

    def filter_fields(self, docs: Union[Dict, List[Dict]]) -> List[Dict]:
        """Filter only fields that are in the resource schema for a list of documents.

        Args:
            docs (Union[dict, list[dict]]): list of documents or single document.

        Returns:
            list[dict]: list of filtered documents.
        """
        if isinstance(docs, dict):
            docs = [docs]
        return [{k: v
                 for k, v in doc.items() if k in self.schema} for doc in docs]


    @property
    def gui(self):
        return self.panel()

    @property
    def df(self):
        return self.to_dataframe()

    def keys(self):
        for page in self.pages():
            yield from page.keys()

    def values(self):
        for page in self.pages():
            yield from page.values()

    def items(self):
        for page in self.pages():
            yield from page.items()

    def records(self):
        for page in self.pages():
            yield from page.records()

    def get_progress_bar_cb(self, total=100, step=1, desc=None, unit=None):
        pbar = tqdm(total=total, desc=desc, unit=unit)
        def cb(idx):
            pbar.update(step)
        return cb

    def pages(self, start=1, end=None, use_async=True, cb=None):
        if use_async:
            pages = asyncio.run(self.pages_async(start=start, end=end, cb=cb))
            yield from pages
        else:
            page_numbers = self.page_numbers
            if cb is None:
                cb = self.get_progress_bar_cb(total=len(page_numbers)*self.items_per_page, step=self.items_per_page,
                        desc=f"Fetching {self.name.lower().replace('_', '')} documents", unit="docs")
            for idx in page_numbers:
                if idx<start:
                    continue
                if end is not None and idx > end:
                    break
                page = self.get_page(idx, cb=True)
                if not len(page):
                    break
                yield page

    async def pages_async(self, start=1, end=None, cb=None):
        aws = []
        page_numbers = self.page_numbers
        if cb is None:
            cb = self.get_progress_bar_cb(total=len(page_numbers)*self.items_per_page, step=self.items_per_page, 
                desc=f"Fetching {self.name.lower().replace('_', ' ')} documents", unit="docs")
            
        for idx in page_numbers:
            if idx<start:
                continue
            if end is not None and idx > end:
                break
            aws.append(self.get_page_async(idx, cb=cb))
        pages = await asyncio.gather(*aws)
        return pages

    def new_item(self, data={}):
        item = self.item_class(**data)
        self[item._id] = item

    def to_records(self):
        return [item.to_dict() for item in self.values()]

    def to_dataframe(self, start=1, end=None, cb=None, use_async=True):
        df = pd.concat([page.to_dataframe() for page in 
                        self.pages(use_async=use_async, start=start, end=end, cb=cb)])
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df
   
    def to_dask(self, pages=None, persist=False):
        try:
            import dask
        except ImportError:
            raise RuntimeError("Dask is not installed.")

        if pages is None:
            pages = self.page_numbers
        columns = [(k, DASK_TYPE_MAPPING[v["type"]]) for k,v in self.schema.items() 
                                    if k in self.fields and not k.startswith("_")]
        column_types = dict(columns)

        url = self._url
        client_kwargs = self.session.get_client_kwargs()
        if client_kwargs["app"] is not None:
            client_kwargs["app"] = dict(client_kwargs["app"].config)

        def get_data(params):
            import httpx
            if client_kwargs["app"] is not None:
                from eve import Eve
                client_kwargs["app"] = Eve(settings=client_kwargs["app"])
            items = []
            with httpx.Client(**client_kwargs) as client:
                try:
                    resp = client.get(url, params=params,)
                    items = resp.json().get("_items", [])
                except:
                    pass
            data = [{k:column_types[k](v) for k,v in item.items() if k in column_types} for item in items]
            return data
        
        if not self.is_tabular:
            import dask.bag as db
            return db.from_sequence([self.get_page_kwargs(i) for i in pages]).map(get_data).flatten()

        import dask.dataframe as dd
        import pandas as pd
        
        def get_df(params):
            data = get_data(params)
            return pd.DataFrame(data, columns=list(column_types))
        dask_name = str(hash((self.name, )+tuple(self.get_page_kwargs(1).values())))
        dsk = {(dask_name, i-1): (get_df, self.get_page_kwargs(i)) for i in pages}
        
        nitems = self.nitems
        divisions = list(range(0, nitems, self.items_per_page))
        if nitems not in divisions:
           divisions = divisions + [nitems]
        
        df = dd.DataFrame(dsk, dask_name, columns, divisions)
        if persist:
            return df.persist()
        return df

   
    def pull(self, start=1, end=None):
        for idx in itertools.count(start):
            if end is not None and idx > end:
                break
            if not self.pull_page(idx):
                break
    
    def push(self, idxs=None):
        if idxs is None:
            idxs = list(self._cache.keys())
        for idx in idxs:
            self._cache[idx].push()

    def get(self, timeout=None, **params):
        params = {k:v if isinstance(v, str) else json.dumps(v, cls=NumpyJSONENncoder) for k,v in params.items()}
        with self.session.Client(timeout=timeout) as client:
            resp = client.get(self._url, params=params)
        return resp.json()

    async def get_async(self, timeout=None, **params):
        params = {k:v if isinstance(v, str) else json.dumps(v, cls=NumpyJSONENncoder) for k,v in params.items()}
        async with self.session.AsyncClient(timeout=timeout) as client:
            resp = await client.get(self._url, params=params)
        return resp.json()
    
    def post_with_files(self, docs):
        success, failed, errors = [], [], []
        with self.session.Client() as client:
            for doc in docs:
                files = {name: BytesIO(doc.pop(name)) for name in self._file_fields if name in doc}
                data = to_data_dict(doc)
               
                try:
                    resp = client.post(self._url, data=data, files=files)
                    success.append(doc)
                except Exception as e:
                    failed.append(doc)
                    errors.append(str(e))
        
        return success, failed, errors

    def post_batched(self, docs):
        data = json.dumps(docs, cls=NumpyJSONENncoder)
        with self.session.Client() as client:
            resp = client.post(self._url, data=data, )
        success, failed, errors = [], [], []
        for doc, result in zip(docs, resp.json()["_items"]):
            if result["_status"]=="OK":
                success.append(doc)
            else:
                failed.append(doc)
                errors.append(result)
            
        if len(failed):
            success = []
        return success, failed, errors

    def post(self, docs, raise_status=True):
        if len(self._file_fields):
            return self.post_with_files(docs)
        else:
            return self.post_batched(docs)

    def find(self, query={}, projection={}, sort="", max_results=25, page_number=1, timeout=None):
        """Find documents in the remote resource that match a mongodb query.

        Args:
            query (dict, optional): Mongo query. Defaults to {}.
            projection (dict, optional): Mongo projection. Defaults to {}.
            sort (Union[string, list[tuple]], optional): Sorting either string as e.g:
                                 city,-lastname or list as e.g: [("city", 1), ("lastname", -1)]
            max_results (int, optional): Items per page. Defaults to 25.
            page_number (int, optional): page to return if query returns more than max_results.\
                                         Defaults to 1.

        Returns:
            list: requested page documents that match query
        """
        if max_results>1 and not any([bool(v) for v in projection.values()]):
            for name in self._file_fields:
                projection[name] = 0
        resp = self.get(where=query,
                        projection=projection,
                        sort=sort,
                        max_results=max_results,
                        page=page_number,
                        timeout=timeout)
        if "_error" in resp:
            return resp["_error"]

        docs = []
        if "_items" in resp:
            docs = resp["_items"]

        if self._file_fields:
            for doc in docs:
                for k in self._file_fields:
                    try:
                        doc[k] = base64.b64decode(doc[k])
                    except:
                        pass
        return docs

    async def find_async(self, query={}, projection={}, sort="", max_results=25, page_number=1, timeout=None):
        """Find documents in the remote resource that match a mongodb query.

        Args:
            query (dict, optional): Mongo query. Defaults to {}.
            projection (dict, optional): Mongo projection. Defaults to {}.
            sort (Union[string, list[tuple]], optional): Sorting either string as e.g:
                                 city,-lastname or list as e.g: [("city", 1), ("lastname", -1)]
            max_results (int, optional): Items per page. Defaults to 25.
            page_number (int, optional): page to return if query returns more than max_results.\
                                         Defaults to 1.

        Returns:
            list: requested page documents that match query
        """
        if max_results>1 and not any([bool(v) for v in projection.values()]):
            for name in self._file_fields:
                projection[name] = 0
        resp = await self.get_async(where=query,
                        projection=projection,
                        sort=sort,
                        max_results=max_results,
                        page=page_number,
                        timeout=timeout)
        docs = []
        if "_items" in resp:
            docs = resp["_items"]

        if self._file_fields:
            for doc in docs:
                for k in self._file_fields:
                    try:
                        doc[k] = base64.b64decode(doc[k])
                    except:
                        pass
        return docs

    def find_page(self, **kwargs):
        """Same as :meth:`eve_panel.EveResource.find()`, only returns an EvePage instance
        """
        docs = self.find(**kwargs)
        items = [self.make_item(**doc) for doc in docs]
        page_number = kwargs.get("page_number", self.page_number)
        page = EvePage(
            name=f'{self._url.replace("/", ".")} page {page_number}',
            _items={item._id: item
                    for item in items},
            fields=self.fields)
        return page

    async def find_page_async(self, **kwargs):
        """Same as :meth:`eve_panel.EveResource.find()`, only returns an EvePage instance
        """
        docs = await self.find_async(**kwargs)
        items = [self.make_item(**doc) for doc in docs]
        page_number = kwargs.get("page_number", self.page_number)
        page = EvePage(
            name=f'{self._url.replace("/", ".")} page {page_number}',
            _items={item._id: item
                    for item in items},
            fields=self.fields)
        return page

    def find_df(self, **kwargs):
        """Same as :meth:`eve_panel.EveResource.find()`, only returns a pandas dataframe

        """
        page = self.find_page(**kwargs)
        df = page.to_dataframe()
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df

    async def find_df_async(self, **kwargs):
        """Same as :meth:`eve_panel.EveResource.find()`, only returns a pandas dataframe

        """
        page = await self.find_page_async(**kwargs)
        df = page.to_dataframe()
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df

    def find_one(self, query: dict={}, projection: dict={}) -> dict:
        """Find the first document that matches the query,
                 optionally project only given fields.

        Args:
            query (dict, optional): Mongo qury to perform. Defaults to {}.
            projection (dict, optional): Mopngo projection. Defaults to {}.

        Returns:
            Union[dict, None]: If document found, returns it. If not returns None.
        """
        docs = self.find(query=query, projection=projection, max_results=1)
        if docs:
            return docs[0]

    async def find_one_async(self, query: dict={}, projection: dict={}) -> dict:
        """Find the first document that matches the query,
                 optionally project only given fields.

        Args:
            query (dict, optional): Mongo qury to perform. Defaults to {}.
            projection (dict, optional): Mopngo projection. Defaults to {}.

        Returns:
            Union[dict, None]: If document found, returns it. If not returns None.
        """
        docs = await self.find_async(query=query, projection=projection, max_results=1)
        if docs:
            return docs[0]

    def find_one_item(self, **kwargs):
        doc = self.find_one(**kwargs)
        if doc:
            return self.make_item(**doc)

    async def find_one_item_async(self, **kwargs):
        doc = await self.find_one_async(**kwargs)
        if doc:
            return self.make_item(**doc)

    def validate_documents(self, docs: List[Dict], coerce=True) -> Tuple:
        """Validates documents against resource schema

        Args:
            docs (list[dict]): list of documents to insert

        Returns:
            tuple[list,list,list]: tuple of documents lists: (valid, rejected, errors) 
        """
        schema = {
            name:
            {k: v
             for k, v in field.items() if k in SUPPORTED_SCHEMA_FIELDS}
            for name, field in self.schema.items()
        }
        for name, sch in schema.items():
            if sch["type"] in TYPE_MAPPING:
                sch["type"] = TYPE_MAPPING[sch["type"]]
        if coerce:
            for sch in schema.values():
                if sch["type"] in COERCERS:
                    sch["coerce"] = COERCERS[sch["type"]]
        v = Validator(schema)
        valid = []
        rejected = []
        errors = []
        for doc in docs:
            if v.validate(doc):
                valid.append(v.document)
            else:
                rejected.append(doc)
                errors.append(v.errors)
        return valid, rejected, errors


    def insert_documents(self, docs: Union[List, Tuple, Dict], validate=True, dry=False) -> Tuple:
        """Insert documents into the database

        Args:
            docs (list): Documents to insert.
            validate (bool, optional): whether to validate schema of docs locally. Defaults to True.
            dry (bool, optional): Enable dry run, will validate but not insert documents into DB. Defaults to False.

        Raises:
            TypeError: raised if docs is not the correct type.

        Returns:
            tuple[list, list, list]: Successfuly inserted, rejected, rejection reasons.
        """
        if isinstance(docs, dict):
            docs = [docs]
        elif isinstance(docs, (tuple,list)):
            docs = list(docs)
        else:
            raise TypeError("Documents must be list/tuple or dict")
        if validate:
            docs, rejected, errors = self.validate_documents(docs)
        else:
            rejected, errors = [], []
        if dry:
            success = docs
        elif docs:
            success, post_rejected, post_errors = self.post(docs)
            rejected.extend(post_rejected)
            errors.extend(post_errors)
        else:
            success = []
        return success, rejected, errors

    def insert_items(self, items: Union[EveItem, List[EveItem]]):
        if isinstance(items, EveItem):
            items = [items]
        docs = [item.to_dict() for item in items]
        return self.insert_documents(self, docs, validate=False)

    def clear_upload_buffer(self):
        """Discards the current upload buffer.
        """
        self.paste_bin.value = ""

    def flush_buffer(self, dry=False):
        """Attempts to insert docs in current upload buffer.

        Args:
            dry (bool, optional): Only validate docs with actually inserting them to DB.
                                Defaults to False.

        Returns:
            list: List of succesfully inserted documents.
        """
        try:
            docs = json.loads(self.paste_bin.value)
        except:
            self.upload_errors = ["Cannot read data. is it valid json?"]
            return []

        if isinstance(docs, dict):
            docs = [docs]
        success, rejected, errors = self.insert_documents(docs, dry=dry)
        if not dry:
            self.paste_bin.value = json.dumps(rejected)
        self.upload_errors = [str(err) for err in errors]
        return success

    def pull_page(self, idx=0, cache_result=True, timeout=None):
        if not idx and cache_result:
            self._cache[idx] = PageZero()
            return False
        page = self.find_page(query=self.filters,
                              projection=self.projection,
                              sort=",".join(self.sorting),
                              max_results=self.items_per_page,
                              page_number=idx,
                              timeout=timeout)
        if page._items and cache_result:
            self._cache[idx] = page
        return page

    async def pull_page_async(self, idx=0, cache_result=True, timeout=None):
        if not idx and cache_result:
            self._cache[idx] = PageZero()
            return False
        page = await self.find_page_async(query=self.filters,
                              projection=self.projection,
                              sort=",".join(self.sorting),
                              max_results=self.items_per_page,
                              page_number=idx,
                              timeout=timeout)
        if page._items and cache_result:
            self._cache[idx] = page
        return page

    def get_page_kwargs(self, idx):
        kwargs =  dict(
            where=self.filters,
            projection=self.projection,
            sort=",".join(self.sorting),
            max_results=self.items_per_page,
            page=idx
        )
        kwargs = {k:v if isinstance(v, str) else json.dumps(v) for k,v in kwargs.items()}
        return kwargs

    def push_page(self, idx):
        if not idx in self._cache or len(self._cache[idx]):
            return
        self._cache[idx].push()

    def get_page(self, idx):
        if idx not in self._cache or not len(self._cache[idx]):
            self.pull_page(idx)
        return self._cache.get(
            idx, EvePage(name="Place holder", fields=self.fields))

    async def get_page_async(self, idx, cb=None, timeout=None):
        if idx not in self._cache or not len(self._cache[idx]):
            await self.pull_page_async(idx, timeout=timeout)
        
        page = self._cache.get(
            idx, EvePage(name="Place holder", fields=self.fields))
        if len(page) and callable(cb):
            cb(idx)
        return page

    def get_page_records(self, idx):
        return self.get_page(idx).to_records()

    async def get_page_records_async(self, idx):
        records = await self.get_page_async(idx).to_records()
        return records

    def get_page_df(self, idx, fields=None):
        df = self.get_page(idx).to_dataframe()
        return df

    async def get_page_df_async(self, idx, fields=None):
        df = await self.get_page_async(idx).to_dataframe()
        return df

    def increment_page(self):
        self.page_number = self.page_number + 1

    def next_page(self):
        self.increment_page()
        return self.current_page()

    def current_page(self):
        return self.get_page(self.page_number)

    def decrement_page(self):
        if self.page_number > 1:
            try:
                self.page_number = self.page_number - 1
            except:
                pass

    def previous_page(self):
        self.decrement_page()
        return self.current_page()

    @param.depends("items_per_page",
                   "filters",
                   "fields",
                   "sorting",
                   watch=True)
    def clear_cache(self):
        self._cache = EvePageCache()
        self._plot = None

    def reload_page(self, page_number=None):
        if page_number is None:
            page_number = self.page_number
        if page_number in self._cache:
            self._cache.pop(page_number)
        self.pull_page(page_number)
        self._cache = self._cache

    def remove_item(self, _id: str) -> bool:
        """Remove a single item from the database.

        Args:
            _id (str): _id field of the item to be removed.

        Returns:
            bool: Whether item was removed succesfully.
        """
        return self[_id].delete()
    
    def remove_items(self, *ids):
        for _id in ids:
            self.remove_item(_id)

    def filter(self, **filters):
        """Filter resource, filters can be any valid mongodb query parameters.

        Returns:
            EveResource: Filtered resource.
        """
        filtered = self.clone(filters=filters)
        filtered.clear_cache()
        return filtered

    def project(self, *fields, **projection):
        """Project resource (only fetch some of the fields.)

        Raises:
            ValueError: raised if passed inconsistent projections.

        Returns:
            EveResource: Projected resource.
        """
        projections ={col:1 for col in fields if col in self.schema}
        projections.update(projection)
        value_sum = sum(list(projections.values()))
        if value_sum==0:
            fields = [c for c in self.schema if c not in projections]
        elif value_sum==len(projections):
            fields = list(projections)
        else:
            raise ValueError("Mongo projections can either be inclusive or exclusive but not both.")
        projected =  self.clone(fields=fields)
        # projected.clear_cache()
        return projected

    def sort(self, *fields):
        """Sort data by given fields,
                     fields starting with `-` will be sorted in descending order.
        """
        sorted_ = self.clone(sorting=[col for col in fields if col in self.sorting_options])
        # sorted_.clear_cache()
        return sorted_

    def paginate(self, page_size: int):
        """Change how many items will be in each page.

        Args:
            page_size (int): number of items per page

        Returns:
            EveResource: Paginated EveResource
        """
        paginated = self.clone(items_per_page=page_size)
        
        return paginated

    def pprint_schema(self):
        """Pretty print the schema in nice yaml format 
        """
        print(yaml.dump(self.schema))

    def clone(self, **kwargs):
        c = super().clone(**kwargs)
        c.clear_cache()
        return c

    @property
    def plot(self):
        """
        Shamelessly copied from the amazing Intake package and modified slightly.
        Returns a hvPlot object to provide a high-level plotting API. Will use Dask
        if available, reading parralelism is set by page size and then data is repartitioned by number 
        of cpus available. small collections are pre-read and converted to regular pandas dataframe.
        To display in a notebook, be sure to run ``panel_eve.output_notebook()``
        first.
        """
        if not self.is_tabular:
            raise TypeError("Plotting API currently only supports tabular data.")

        if self._plot is None:
            try:
                from hvplot import hvPlot
            except ImportError:
                raise ImportError("The eve_panel plotting API requires hvplot."
                                "hvplot may be installed with:\n\n"
                                "`conda install -c pyviz hvplot` or "
                                "`pip install hvplot`.")
            nitems = self.nitems
            if nitems<self.items_per_page:
                data = self.df
            else:
                try:
                    import dask
                    persist = (nitems/self.items_per_page)<mp.cpu_count()
                    data = self.to_dask(persist=persist)
                    data = data.repartition(npartitions=min(data.npartitions, mp.cpu_count()))
                    if persist and nitems<1e4:
                        data = data.compute()
                except ImportError:
                    data = self.df
            metadata = self.metadata.get('plot', {})
            fields = self.metadata.get('fields', {})
            for attrs in fields.values():
                if 'range' in attrs:
                    attrs['range'] = tuple(attrs['range'])
            metadata['fields'] = fields
            plots = self.metadata.get('plots', {})
            self._plot = hvPlot(data, custom_plots=plots, **metadata)
        return self._plot

    @property
    def plots(self):
        return list(self.metadata.get('plots', {}))

    @param.depends("_url")
    def upload_view(self):
        clear_button = pn.widgets.Button(name="Clear buffer",
                                         button_type="warning",
                                         width_policy='max',
                                         sizing_mode=self.sizing_mode,
                                         width=int(self.max_width/4),
                                         max_width=int(self.max_width/3),
                                         )
        clear_button.on_click(lambda event: self.clear_upload_buffer())

        upload_file = pn.widgets.FileInput(accept=",".join(
            [f".{ext}" for ext in FILE_READERS]),
                                           width=int(self.max_width/4),
                                           max_width=int(self.max_width/3),
                                           )
        upload_file.link(self, filename="_filename", value="_file_buffer")
        upload_file_button = pn.widgets.Button(name="Read file",
                                               button_type="primary",
                                               width_policy='max',
                                               sizing_mode=self.sizing_mode,
                                               width=int(self.max_width/4),
                                               max_width=int(self.max_width/3),
                                               )
        upload_file_button.on_click(lambda event: self._read_file_buffer())

        upload_button = pn.widgets.Button(name="Insert to DB",
                                          button_type="success",
                                          width_policy='max',
                                          sizing_mode=self.sizing_mode,
                                          width=int(self.max_width/4),
                                          max_width=int(self.max_width/3),
                                          )
        upload_button.on_click(lambda event: self.flush_buffer())
        read_clipboard_button = pn.widgets.Button(name="Read Clipboard",
                                                  button_type="primary",
                                                  width_policy='max',
                                                  sizing_mode=self.sizing_mode,
                                                  width=int(self.max_width/4),
                                                  max_width=int(self.max_width/3),
                                                  )
        read_clipboard_button.on_click(lambda event: self.read_clipboard())

        validate_button = pn.widgets.Button(name="Validate",
                                                  button_type="primary",
                                                  width_policy='max',
                                                  sizing_mode=self.sizing_mode,
                                                  width=int(self.max_width/4),
                                                  )
        validate_button.on_click(lambda event: self.flush_buffer(dry=True))

        first_row_buttons = pn.Row(upload_file, upload_file_button, read_clipboard_button)
        second_row_buttons = pn.Row(validate_button, clear_button, upload_button)
        input_buttons = pn.Column(first_row_buttons, second_row_buttons)
        upload_view = pn.Column(self.paste_bin, input_buttons,"### Errors:",  self.upload_errors_view)
        return upload_view

    @param.depends("page_number", "_cache", "_page_view_format")
    def current_page_view(self):
        page = self.get_page(self.page_number)
        if page is None:
            return pn.panel(f"## No data for page {self.page_number}.")
        return getattr(page,
                       self._page_view_format.lower() + "_view", page.panel)()

    @param.depends("upload_errors")
    def upload_errors_view(self):
        alerts = [
            pn.pane.Alert(err, alert_type="danger",
            margin=2,
            width_policy='max',
            sizing_mode=self.sizing_mode,
            width=self.max_width-30,
            max_width=self.max_width,
            )
            for err in self.upload_errors
        ]
        return pn.Column(*alerts,
                         scroll=True,
                         max_height=300,
                         width_policy='max',
                         sizing_mode=self.sizing_mode,
                         max_width=self.max_width-30)

    @param.depends("current_page")
    def download_view(self):
        # data_selection = pn.widgets.RadioBoxGroup(options=["Current Page", "Range"])
        # page_range = pn.widgets.IntRangeSlider()
        page_num_select = pn.widgets.IntInput(name="Page", value=self.page_number, step=1, start=1, end=int(1e6))
        def cb():
            page = self.get_page(page_num_select.value)
            return page.to_file()
        download_button = pn.widgets.FileDownload(callback=cb, label="Download JSON", filename=f'{self.name}_page{page_num_select.value}.json')
        return pn.Column(page_num_select, download_button )

    @param.depends("_plot_selection")
    def selected_plot_view(self):
        if self._plot_selection in self.plots:
            plot = getattr(self.plot, self._plot_selection)()
            return pn.panel(plot)
        else:
            return pn.Column("# No plot selected.")

    def make_panel(self, show_client=True, tabs_location='above'):
        plot_selector = pn.Param(self.param._plot_selection,
        widgets={"_plot_selection":{"type": pn.widgets.Select,
                                    "options": ["None"]+self.plots}},
                                width_policy='max',
                                sizing_mode='stretch_width',
                                width=int(self.max_width/4),
                                max_width=int(self.max_width/3),
                                )
        plotting = pn.Column(plot_selector,
                             self.selected_plot_view,
                             width_policy='max',
                             sizing_mode='stretch_width',
                             max_width=int(self.max_width - 30),
                             )

        column_select = pn.Param(self.param.fields,
                            width_policy='max',
                            sizing_mode='stretch_width',
                            max_width=int(self.max_width - 30),
                            widgets={
                                "fields": {
                                    "type": pn.widgets.MultiChoice,
                                    "options": self.field_options,
                                    "width": int(self.max_width - 30)
                                }
                            })

        sorting_select = pn.Param(self.param.sorting,
                            width_policy='max',
                            sizing_mode='stretch_width',
                            max_width=int(self.max_width - 30),
                            widgets={
                                "sorting": {
                                    "type": pn.widgets.MultiChoice,
                                    "options": self.sorting_options,
                                    "width": int(self.max_width - 30)
                                }
                            })

        page_settings = pn.Column(pn.Row(self.param.items_per_page,
                                         self.param.filters,
                                         self.param._page_view_format,
                                         width_policy='max',
                                         sizing_mode=self.sizing_mode,
                                         max_width=int(self.max_width - 50)),
                                  column_select,
                                  sorting_select,
                                  max_width=int(self.max_width - 10))
        if show_client:
            page_settings.append("### HTTP client")
            page_settings.append(pn.layout.Divider())
            page_settings.append(self.session.panel)


        tabs = pn.Tabs(
                ("Data", self.current_page_view),
                ("Plots", plotting),
                ("Upload", self.upload_view()),
                ("Download", self.download_view),
                ("Config", page_settings),
                # ("Errors", self.session.messages),
                dynamic=False,
                tabs_location=tabs_location,
                width_policy='max',
                sizing_mode=self.sizing_mode,
                max_width=self.max_width-20,
            )

        buttons = pn.Param(self.param,
                           parameters=[
                               "_prev_page_button", "page_number",
                               "_reload_page_button", "_next_page_button",
                               "_clear_cache_button",
                           ],
                           
                           default_layout=pn.Row,
                           name="",
                           width_policy='max',
                           sizing_mode=self.sizing_mode,
                           width=int(self.max_width/2),
                           min_width=300,
                           max_width=int(self.max_width/1.5))

        header = pn.Row(
            f'## {self.name.replace("_", " ").title()} resource',
            pn.Spacer(sizing_mode='stretch_both'),
            pn.Column(pn.Spacer(), self.gui_progress, pn.Spacer()),
           
            buttons,
            pn.Spacer(sizing_mode='stretch_both'),
            align="center",
            
        )
        # if settings.SHOW_INDICATOR:
        #     header.append(self.session.busy_indicator)

        view = pn.Column(
            header, 
            tabs,
            width_policy='max',
            sizing_mode=self.sizing_mode,
            width=self.max_width-20,
            max_width=self.max_width,
            max_height=self.max_height,
            )

        return view
    


import base64

import panel as pn
import param
import pkg_resources

import httpx
import webbrowser
import time
import secrets
import getpass
import authlib
from authlib import jose
from authlib.jose.rfc7517.models import KeySet

from contextlib import contextmanager, asynccontextmanager

from eve_panel import config


class OidcKeySet(param.Parameterized):
    oauth_domain = param.String(config.OAUTH_DOMAIN)
    cert_path = param.String(config.OAUTH_CERT_PATH)

    _keyset = param.ClassSelector(KeySet, default=KeySet({}))

    _keys_timestamp = param.Number(0)
    _keys_ttl = param.Number(300)

    def fetch_keys(self, headers={}):
        with httpx.Client(base_url=self.oauth_domain, headers=headers) as client:
            r = client.get(self.cert_path)
            r.raise_for_status()
        keys = r.json()
        self._keyset = jose.JsonWebKey.import_key_set(keys)

    def extract_claims(self, token):
        header_str = authlib.common.encoding.urlsafe_b64decode(token.split(".")[0].encode()).decode('utf-8')
        header = authlib.common.encoding.json_loads(header_str)
        key = self.find_by_kid(header["kid"])
        return jose.jwt.decode(token, key)

    def extract_verified_claims(self, token, options={}):
        try:
            claims = self.extract_claims(token)
            claims.options = options
            claims.validate()
            return claims
            
        except Exception as e:
            logger.error(f"Exception raised while validating claims: {e}")
            return jose.JWTClaims("", "")

    def validate_claims(self, token, **required_claims):
        options = {k: {"value": v, "essential": True} for k,v in required_claims.items()}
        claims = self.extract_verified_claims(token, options)
        claims.validate()

    def find_by_kid(self, kid):
        if kid not in self._keyset.keys:
            self.fetch_keys()
        return self._keyset.find_by_kid(kid)

    def __getitem__(self, kid):
        return self.find_by_kid(kid)



class OidcToken(param.Parameterized):
    client_id = param.String()
    access_token = param.String()
    id_token = param.String()
    refresh_token = param.String()
    expires = param.Number()
    scope = param.String()
    token_type = param.String("Bearer")

    @property
    def claims(self):
        claims = self.extract_claims(self.access_token)
        try:
            claims.validate()
        except:
            return {}
        return dict(claims)

    @property
    def expired(self):
        return time.time()>self.expires

    @classmethod
    def from_file(cls, path):
        with open(path, "r") as f:
            data = json.load(f)
        return cls(**data)

    def to_dict(self):
        return {k:v for k,v in self.param.get_param_values() if not k.startswith("_")}
        
    def to_file(self, path):
        with open(path, "w") as f:
            json.dump(self.to_dict(), f)

    def refresh_token(self, oauth_domain, oauth_token_path, rtoken, headers={}):
        with httpx.Client(base_url=oauth_domain, headers=headers) as client:
            r = client.post(
                oauth_token_path,
            headers={"content-type":"application/x-www-form-urlencoded"},
            data={
                "grant_type": "refresh_token",
                "refresh_token": rtoken,
                "client_id": self.client_id,
            }
            )
            r.raise_for_status()
            params = r.json()
            params["expires"] = time.time() + params.pop("expires_in", 1e6)
            self.set_param(**params)           
        

class OidcAuthFlow(param.Parameterized):
    client_id = param.String()
    device_code = param.String()
    user_code = param.String()
    verification_uri = param.String()
    expires = param.Number()
    interval = param.Number(5)
    verification_uri_complete = param.String()
    
    def request_token(self, oauth_domain, oauth_code_path, client_id, scope, audience, headers={}):
        with httpx.Client(base_url=oauth_domain, headers=headers) as client:
            r = client.post(
                oauth_code_path,
                data={
                    "client_id": client_id,
                    "scope": scope,
                    "audience": audience,
                    },
                headers={"content-type": "application/x-www-form-urlencoded"})
            r.raise_for_status()
        self.client_id = client_id
        params = r.json()
        params["expires"] = time.time() + params.pop("expires_in", 1)
        for k,v in params.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def fetch_token(self, oauth_domain, oauth_token_path, headers={}):
        with httpx.Client(base_url=oauth_domain, headers=headers) as client:
            r = client.post(
                oauth_token_path,
            headers={"content-type":"application/x-www-form-urlencoded"},
            data={
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "device_code": self.device_code,
                "client_id": self.client_id,
            }
            )
            r.raise_for_status()
            params = r.json()
            params["expires"] = time.time() + params.pop("expires_in", 1e6)
            params["client_id"] = self.client_id
        return OidcToken(**params)
    


    def await_token(self, oauth_domain, oauth_token_path, headers={}):
        while True:
            if time.time()>self.expires:
                raise TimeoutError("Device code hase expired but not yet authorized.")
            try:
                s = self.fetch_token(oauth_domain, oauth_token_path, headers=headers)
                return s
            except:
                time.sleep(self.interval)

    def perform(self, oauth_domain, oauth_code_path, oauth_token_path,
                 client_id, scope, audience, headers={}, open_browser=False, print_url=False):
        self.request_token(oauth_domain, oauth_code_path, client_id, scope, audience, headers=headers)
        if print_url:
            print(f"Authorization URL: {self.verification_uri_complete}")
        if open_browser:
            webbrowser.open(self.verification_uri_complete)        
        return self.await_token(oauth_domain, oauth_token_path, headers=headers)
        



class OidcSession(param.Parameterized):

    message = param.String("")
    _gui = None
    _cb = None

    oauth_domain = param.String(config.OAUTH_DOMAIN)
    oauth_code_path = param.String(config.OAUTH_CODE_PATH)
    oauth_token_path = param.String(config.OAUTH_TOKEN_PATH)

    client_id = param.String(config.DEFAULT_CLIENT_ID) 
    scopes = param.List([])
    audience = param.String(config.DEFAULT_AUDIENCE)

    flow = param.ClassSelector(OidcAuthFlow, default=OidcAuthFlow())
    keyset = param.ClassSelector(OidcKeySet, default=certs)
    token = param.ClassSelector(OidcToken, default=None)
    state = param.Selector(["Disconnected", "Logged in", "Awaiting token", "Checking token ready", "Token expired"],
                             default="Disconnected")

    @property
    def scope(self):
        scopes = set(config.DEFAULT_SCOPE.split(" ") + self.scopes)
        return " ".join(scopes)

    def login(self, open_browser=False, print_url=False, extra_headers={}):
        self.token = self.flow.perform(self.oauth_domain, self.oauth_code_path, self.oauth_token_path, 
                                        self.client_id, self.scope, self.audience, headers=extra_headers,
                                      open_browser=open_browser, print_url=print_url)
        if self.token:
            self.state = "Logged in"

    def request_token(self,  extra_headers={}):
        
        self.flow.request_token(self.oauth_domain, self.oauth_code_path, 
                                self.client_id, self.scope, self.audience, headers=extra_headers)
        self.state = "Awaiting token"

    def token_ready(self, extra_headers={}):
        if self.token is None:
            try:
                self.token = self.flow.fetch_token(self.oauth_domain,
                                         self.oauth_token_path, headers=extra_headers)
                self.state = "Logged in"
                return True
            except:
                pass
        return False
    
    def refresh_token(self, extra_headers={}):
        self.token = self.flow.refresh_token(self.oauth_domain, self.oauth_token_path, 
                                             self.token.refresh_token, headers=extra_headers)

    def logged_in(self):
        if self.token is None: 
            return False
        if self.token.expired:
            return False
        return True

    @property
    def id_token(self):
        return self.token.id_token

    @property
    def access_token(self):
        return self.token.access_token

    @property
    def profile(self):
        claims = self.keyset.extract_verified_claims(self.id_token)
        return {k:v for k,v in claims.items() if k not in claims.REGISTERED_CLAIMS}
    
    @property
    def claims(self):
        claims = self.keyset.extract_verified_claims(self.access_token)
        return {k:v for k,v in claims.items() if k in claims.REGISTERED_CLAIMS}

    @property
    def extra_claims(self):
        claims = self.keyset.extract_verified_claims(self.access_token)
        return {k:v for k,v in claims.items() if k not in claims.REGISTERED_CLAIMS}

    @property
    def permissions(self):
        claims = self.keyset.extract_verified_claims(self.access_token)
        return claims.get("permissions", [])

    @contextmanager
    def Client(self, *args, **kwargs):
        kwargs["headers"] = kwargs.get("headers", {})
        kwargs["headers"]["Authorization"] = f"Bearer {self.access_token}"
        
        client = httpx.Client(*args, **kwargs)
        try:
            yield client
        finally:
            client.close()

    @asynccontextmanager
    async def AsyncClient(self, *args, **kwargs ):
        kwargs["headers"] = kwargs.get("headers", {})
        kwargs["headers"]["Authorization"] = f"Bearer {self.access_token}"
        client = httpx.AsyncClient(*args, **kwargs)
        try:
            yield client
        finally:
            await client.aclose()

    def authorize(self):
        webbrowser.open(self.flow.verification_uri_complete)
        

    @property
    def gui(self):
        if self._gui is None:
            self._gui = pn.panel(self._make_gui)
        return self._gui

    def await_token_cb(self):
        if self.token_ready() and self.token:
            self._cb.stop()
        
    def login_requested(self, event):
        try:
            self.request_token()
            logger.info("Sent request...")
            self._cb = pn.state.add_periodic_callback(self.await_token_cb,
                                                    1000*self.flow.interval,
                                                    timeout=1000*max(1, int(self.flow.expires-time.time())))
        except Exception as e:
            logging.error(e)
            print(e)

    def logged_in_gui(self):
        profile = self.profile
        details = pn.Row(
            pn.pane.PNG(profile.get("picture", config.DEFAULT_AVATAR), width=60, height=60),
            pn.widgets.TextInput(name='Name', value=profile.get("name", "Unknown"), disabled=True, height=35),
            pn.widgets.TextInput(name='Email', value=profile.get("email", "Unknown"), disabled=True, height=35),
            
            height=70
        )
        token = pn.widgets.input.TextAreaInput(name='Access token', value=self.access_token, width=700, height=100)
        
        token_props = pn.Row(
            pn.widgets.TextInput(name='Scope', value=self.token.scope, disabled=True, height=35),
            pn.widgets.DatetimeInput(disabled=True, name="Expiration date",
                         value=datetime.utcfromtimestamp(self.token.expires)),
            
            width=700,
        )
        return pn.Column(
                    details,
                    token,
                    token_props,
                    height=300,
                    width=700)

    def awaiting_token_gui(self):
        return pn.Column(url_link_button(self.flow.verification_uri_complete))
    
    def token_expired_gui(self):
        refresh = pn.widgets.Button(name="Renew", align="end")
        refresh.on_click(lambda e: self.refresh_token())
        return refresh

    @param.depends("state")
    def _make_gui(self):
        status = pn.indicators.BooleanStatus(width=15, height=15, value=True, color="danger")
        header = pn.Row(status, f"Status: {self.state}.")
        panel = pn.Column(header)
        if self.state == "Logged in":
            status.color = "success"
            panel.append(self.logged_in_gui())
            
        elif self.state == "Awaiting token":
            status.color = "primary"
            panel.append(self.awaiting_token_gui())
        else:
            login = pn.widgets.Button(name="Login", button_type='primary', width=30)
            login.on_click(self.login_requested)
            panel.append(login)
        return panel

    def _repr_mimebundle_(self, include=None, exclude=None):
        return self.gui._repr_mimebundle_(include=include, exclude=exclude)




"""
auth module
====================================
Authentication and Authorization handling
"""
import pkg_resources

from .base import EveAuthBase, EveNoAuth
from .basic_auth import EveBasicAuth
from .bearer import EveBearerAuth
from .oauth2 import Oauth2DeviceFlow

AUTH_CLASSES = {
    None: EveNoAuth,
    "Basic": EveBasicAuth,
    "Bearer": EveBearerAuth,
    "Oauth2 Device Flow": Oauth2DeviceFlow,
}

for entry_point in pkg_resources.iter_entry_points('eve_panel.auth'):
    AUTH_CLASSES[entry_point.name] = entry_point.load()

import pkg_resources

DEFAULT_AUTH = None



class EvePanelError(Exception):
    pass

class AuthError(EvePanelError):
    pass

class ServerError(EvePanelError):
    def __init__(self, status_code, message=""):
        self.status_code = status_code
        super().__init__(message)



import panel as pn
import param
from bson import ObjectId
import json
from io import BytesIO
import logging

from .settings import config as settings
from .eve_model import DefaultLayout, EveModelBase
from .field import EveField
from .session import EveSessionBase
from .types import TYPE_MAPPING
from .widgets import get_widget
from .utils import NumpyJSONENncoder, to_data_dict, to_json_compliant

logger = logging.getLogger(__name__)

class EveItem(EveModelBase):
    session = param.ClassSelector(EveSessionBase, constant=True, precedence=-1)
    _resource_url = param.String(precedence=-1)
    _etag = param.String(allow_None=True, precedence=15)
    _version = param.Integer(default=1,
                             bounds=(1, None),
                             allow_None=True,
                             precedence=13)
    _latest_version = param.Integer(default=1,
                                    bounds=(1, None),
                                    allow_None=True,
                                    constant=True,
                                    precedence=14)

    _save = param.Action(lambda self: self.save(), label="Save", precedence=16)
    _delete = param.Action(lambda self: self.delete(),
                           label="Delete",
                           precedence=16)
    _delete_requested = param.Boolean(False, precedence=-1)
    _deleted = param.Boolean(False, precedence=-1)
    _verification = param.String("Enter item id here to enable deletion.", label="", precedence=1)
    _clone = param.Action(lambda self: self.clone(),
                          label="Clone",
                          precedence=16)
    _reload = param.Action(lambda self: self.pull(),
                          label="Reload",
                          precedence=16)

    def __init__(self, **params):
        params["_id"] = params.get("_id", str(ObjectId()))
        if "name" not in params:
            params["name"] = f'{self.__class__.__name__}_{params["_id"]}'
        params = {k: v for k, v in params.items() if hasattr(self, k)}
        super().__init__(**params)

    @classmethod
    def from_schema(cls,
                    name,
                    schema,
                    resource_url,
                    session=None,
                    data={}):
        
        params = dict(
            schema=param.Dict(default=schema,
                               allow_None=False,
                               constant=True,
                               precedence=-1),
            _resource_url=param.String(default=resource_url, precedence=-1),
        )
        _widgets = {
            "_etag": {
                "type": pn.widgets.TextInput,
                "disabled": True
            },
            "_version": {
                "type": pn.widgets.IntInput,
                "disabled": False
            }
        }

        for field_name, field_schema in schema.items():
            kwargs = {"precedence": 10}
            if field_name == "_id":
                kwargs["precedence"] = 1
            extended_name = f"{name.title()}{field_name.title()}"

            if "allowed" in field_schema:
                class_ = param.Selector
                kwargs["objects"] = field_schema["allowed"]

            elif field_schema["type"] in TYPE_MAPPING:
                class_ = EveField(extended_name, field_schema,
                                  TYPE_MAPPING[field_schema["type"]])
            else:
                continue
            if "default" in field_schema:
                kwargs["default"] = field_schema["default"]
            else:
                kwargs["default"] = None

            widget = get_widget(extended_name, field_schema)
            if widget is not None:
                _widgets[field_name] = widget

            if field_schema.get("required", False):
                kwargs["allow_None"] = False

            bounds = (field_schema.get("min",
                                       None), field_schema.get("max", None))
            if any(bounds):
                kwargs["bounds"] = bounds
            kwargs["readonly"] = field_schema.get("readonly", False)
            params[field_name] = class_(**kwargs)
        params["_widgets"] = param.Dict(default=_widgets,
                                        constant=True,
                                        precedence=-1)
        klass = type(name, (EveItem, ), params)

        return klass(schema=schema,
                     _widgets=_widgets,
                     _resource_url=resource_url,
                     session=session,
                     **data)

    @property
    def url(self):
        return "/".join([self._resource_url, self._id])

    def save(self):
        self.push()

    def to_record(self, exclude_files=True):
        obj = {}
        for k in self.schema:
            v = getattr(self, k)
            if exclude_files and isinstance(v, bytes):
                continue
            obj[k] = v
        return obj

    def to_json(self):
        obj = self.to_record()
        return json.dumps(obj, cls=NumpyJSONENncoder)

    def to_dict(self):
        return self.to_record()

    def keys(self):
        yield from self.to_record().keys()

    def values(self):
        yield from self.to_record().values()

    def items(self):
        yield from self.to_record().items()

    def __getitem__(self, key):
        if key in self.schema:
            return getattr(self, key)
        else:
            raise KeyError(f"{key} not found.")
    
    def __setitem__(self, key, value):
        if key in self.schema:
            setattr(self, key, value)
        else:
            raise KeyError(f"{key} cannot be set.")

    @param.depends("_version", watch=True)
    def pull(self):
        if self._version is None:
            version = 1
        else:
            version = self._version
        with self.session.Client() as client:
            resp = client.get(self.url, params=dict(version=version))
        data = resp.json()
        if not data:
            return
        for k, v in data.items():
            if hasattr(self, k):
                try:
                    param = getattr(self.param, k)
                    if param.constant or param.readonly:
                        setattr(self, getattr(param, "_internal_name"), v)
                    else:
                        self.param.set_param(k=v)
                except Exception as e:
                    logger.error(str(e))
                    pass
    
    def get_version(self, version):
        vers = self.clone(_version=version)
        vers.pull()
        return vers
        
    def all_versions(self):
        with self.session.Client() as client:
            resp = client.get(self.url, params=dict(version='all'))
            resp.raise_for_status()
        data = resp.json()
        if not data:
            return []
        return [
            self.__class__(**doc,
                           session=self.session,
                           _resource_url=self._resource_url)
            for doc in data.get("_items", [])
        ]

    def version_diffs(self):
        with self.session.Client() as client:
            resp = client.get(self.url, params=dict(version='diffs'))

        data = resp.json()
        if not data:
            return []
        return [
            self.__class__(**doc,
                           session=self.session,
                           _resource_url=self._resource_url)
            for doc in data.get("_items", [])
        ]

    def push(self):
        headers ={}
        if self._version == self._latest_version:
            headers["If-Match"] = self._etag
        data = {k: getattr(self, k) for k in self.schema if not k.startswith("_")}
        doc = {k:v for k,v in data.items() if v is not None}
        files = {name: BytesIO(doc.pop(name)) for name, value in data.items()
                    if isinstance(value, bytes)}
        data = to_data_dict(doc)
        
        with self.session.Client() as client:
            resp = client.put(self.url, data=data, files=files, headers=headers)
        self.pull()

    def patch(self, *fields):
        headers ={}
        if self._version == self._latest_version:
            headers["If-Match"] = self._etag
            
        data = {}
        for k in fields:
            data[k] = getattr(self, k)
        doc = {k:v for k,v in data.items() if v is not None}
        files = {name: BytesIO(doc.pop(name)) for name, value in data.items() if isinstance(value, bytes)}
        data = to_data_dict(doc)

        with self.session.Client() as client:
            resp = client.patch(self.url, data=data, files=files, headers=headers)
            resp.raise_for_status()
        self.pull()

    def delete(self, verification=None):
        if verification is not None and verification != self._id:
            print(verification)
            return False
        headers ={}
        if self._version == self._latest_version:
            headers["If-Match"] = self._etag
        with self.session.Client() as client:
            resp = client.delete(self.url, headers=headers)
            self._deleted = True
        return self._deleted

    # def clone(self, **kwargs):
    #     data = {k: getattr(self, k) for k in self.schema}
    #     data.update(kwargs)
    #     return self.__class__(**data)

    @param.depends("_delete_requested")
    def buttons(self):
        param_buttons = pn.Param(self.param,
                           parameters=["_reload", "_save"],
                           widgets={
                               "_reload": {
                                   "type": pn.widgets.Button,
                                   "button_type": "primary"
                               },
                               "_save": {
                                   "type": pn.widgets.Button,
                                   "button_type": "success"
                               },
                           },
                           show_name=False,
                           default_layout=pn.Row)

        delete_button = pn.widgets.Button(name="Remove", align="center", max_width=50)
        verification = pn.widgets.TextInput(placeholder="Enter item id here to enable deletion.")
        row = pn.Row(param_buttons, verification, delete_button)
        if self._deleted:
            delete_button.name = "Deleted."
            delete_button.disabled = True
            row.pop(1)
        elif self._delete_requested:
            delete_button.name = "Delete"
            delete_button.button_type = "danger"
            def onclick(event):
                delete_button.disabled = True
                self.delete(verification=verification.value)
                delete_button.disabled = False
            delete_button.on_click(onclick)
        else:
            delete_button.disabled = False
            delete_button.button_type = "warning"
            def onclick(event):
                self._delete_requested = True
            delete_button.on_click(onclick)
            row.pop(1)
        return row

    def make_panel(self):
        header = pn.Column(
            pn.layout.Divider(),
            f"### {self.name}",
            width_policy='max',
            sizing_mode=self.sizing_mode,
            width=self.max_width,
        )
        
        # buttons = pn.Row(param_buttons, self.delete_button)
    
        editors = pn.Param(self.param,
                           show_name=False,
                           default_layout=DefaultLayout,
                           widgets=self._widgets,
                           parameters=list(self.schema)+settings.META_FIELDS,
                           width_policy='max',
                           sizing_mode=self.sizing_mode,
                           width=self.max_width,
                           )
        # media_previews = 
        return pn.Column(header,
                        editors,
                        self.buttons,
                        width_policy='max',
                        sizing_mode=self.sizing_mode,
                        width=self.max_width,
                        )

    def __repr__(self):
        return f"{self.__class__.__name__}(_id={self._id or self.name})"



import param
from bson import objectid
import numpy as np
import pandas as pd
import base64

class CoerceClassSelector(param.ClassSelector):
    def __set__(self, obj, val):
        try:
            val = self.class_(val)
        except:
            pass
        super().__set__(obj, val)


def objectid_param(**kwargs):
    return CoerceClassSelector(str, constant=True, **kwargs)

def bytes_param(**kwargs):
    return param.ClassSelector(bytes, **kwargs)

def set_param(**kwargs):
    return param.ClassSelector(set, **kwargs)


TYPE_MAPPING = {
    "objectid": objectid_param,
    "boolean": param.Boolean,
    "binary": bytes_param,
    "date": param.Date,
    "datetime": param.Date,
    "dict": param.Dict,
    "float": param.Number,
    "integer": param.Integer,
    "list": param.List,
    "number": param.Number,
    "set": set_param,
    "string": param.String,
    "media": bytes_param,
}

DASK_TYPE_MAPPING = {
    "objectid": str,
    "boolean": bool,
    "binary": bytes,
    "date": np.datetime64,
    "datetime": np.datetime64,
    "dict": dict,
    "float": float,
    "integer": int,
    "list": list,
    "number": float,
    "set": set,
    "string": str,
    "media": bytes,

}

def to_binary(x):
    if isinstance(x, str):
        x = str.encode(x)
    return x

def base64_to_binary(x):
    if isinstance(x, str):
        x = base64.b64decode(x)
    return x

COERCERS = {
    "objectid": str,
    "boolean": bool,
    "binary": to_binary,
    "date": pd.to_datetime,
    "datetime": pd.to_datetime,
    "dict": dict,
    "float": float,
    "integer": int,
    "list": list,
    "number": float,
    "set": set,
    "string": str,
    "media": base64_to_binary,
}







import base64
import panel as pn
import param
import getpass
import logging

from .base import EveAuthBase

logger = logging.getLogger(__name__)


class EveBearerAuth(EveAuthBase):
    """
    Support for Eve bearer auth.

    Inheritance:
        EveAuthBase:

    """
    token = param.String(doc="Beaer token")

    def get_headers(self):
        if self.token:
            return {"Authorization": f"Bearer {self.token}"}
        else:
            return {}

    def login(self, token=None):
        if token is None:
            token = input("Access token: ")
        self.token = token
        return bool(self.token)

    def logout(self):
        self.token = ""
        
    @property
    def logged_in(self):
        return bool(self.token)

    def make_panel(self):
        return pn.panel(self.param.token,
                        max_width=self.max_width,
                        max_height=self.max_height,
                        sizing_mode=self.sizing_mode,
        )

    def credentials_view(self):
        return self.panel()


import ast
import json
import base64

import panel as pn
import param
from eve.io.mongo.validation import Validator
from panel.widgets import LiteralInput


class LiteralSchemaInputBase(LiteralInput):
    """[summary]

    Args:
        LiteralInput ([type]): [description]
    """
    def validate_schema(self, value):
        return True

    def _process_property_change(self, msg):
        msg = super(LiteralSchemaInputBase, self)._process_property_change(msg)
        if msg['value'] == self.value:
            return msg
        new_state = ''
        if 'value' in msg:
            value = msg.pop('value')
            if not self.validate_schema(value):
                new_state = ' (invalid)'
                value = self.value
            msg['value'] = value
            msg['name'] = msg.get('title', self.name).replace(
                self._state, '').replace(new_state, '') + new_state
            self._state = new_state
            self.param.trigger('name')
        return msg


def LiteralSchemaInput(name, schema, type_=None):
    validator = Validator({"value": schema})

    def validate_schema(self, value):
        return validator.validate({"value": value})

    params = {
        "validate_schema": validate_schema,
        "type": type_,
    }
    return type(name + "InputWidget", (LiteralSchemaInputBase, ), params)


def PDFViewer(pdf, width=800, height=500):
    if isinstance(pdf, bytes):
        pdf = base64.b64encode(pdf).decode()
    return pn.pane.HTML(f'<iframe width={width} height={height} src="data:application/pdf;base64,{pdf}" type="application/pdf"></iframe>',
                   width=1000,sizing_mode="stretch_both")

def PNGViewer(png, width=800, height=500):
    if isinstance(png, bytes):
        png = base64.b64encode(png).decode()
    src = f"data:image/png;base64,{png}"
    return pn.pane.HTML(f"<img src='{src}' width={width} height={height}></img>")
                
class FileInputPreview(pn.widgets.FileInput):

    @param.depends('value')
    def preview(self):
        if not self.filename:
            return pn.Column()
        if self.filename.endswith(".png"):
            return PNGViewer(self.value)
        elif self.filename.endswith(".pdf"):
            return PDFViewer(self.value)

    def _repr_mimebundle_(self, include=None, exclude=None):
        return pn.Column(self, self.preview)

WIDGET_MAPPING = {
    "media": {"type": pn.widgets.FileInput, "align": "end"},
}


def get_widget(name, schema):
    if schema["type"] == "dict" and "schema" in schema:
        return LiteralSchemaInput(name, schema, dict)
    elif schema["type"] == "list" and "schema" in schema:
        return LiteralSchemaInput(name, schema, list)
    else:
        return WIDGET_MAPPING.get(schema["type"], None)



import numpy as np
import json
from functools import wraps
import re
url_regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

def is_valid_url(url):
    return re.match(url_regex, url) is not None

def requires_login(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.session.logged_in:
            raise RuntimeError("You are not logged in.")
        return method(self, *args, **kwargs)
    return wrapper

def to_json_compliant(obj):
    if isinstance(obj, np.integer):
            return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return obj
class NumpyJSONENncoder(json.JSONEncoder):
    def default(self, obj):
        obj = to_json_compliant(obj)
        return super(NumpyJSONENncoder, self).default(obj)

def to_data_dict(doc):
    data = {}
    for k,v in doc.items():
        if isinstance(v, str):
            data[k] = v
        else:
            data[k] = json.dumps(v, cls=NumpyJSONENncoder)
    return data